#!/usr/local/bin/ruby

class Plain

  def initialize
    @tab = -2
    @col = -2
    @maxcol = 65
    @stac = []
    @in_pre = false
    @emp = 0
  end

  attr :in_pre

  def shift
    @tab+= 2
    @col+= 2
  end

  def unshift
    @tab-= 2
    @col-= 2
  end

  def line_feed
    $_ = fix
    if ! @in_pre && /\w/
      tab = @tab
      tab = 0 if tab < 0
      tab+= 1 if @in_dd
      print '  ' * tab
    end
    if /\w/
      @emp = 0
    else
      @emp+= 1
    end
    $_ << "\e" if @emp == 0 || @emp == 1
    if @in_pre
      gsub!(/\e%/, "\e%%")
      $_ << "%\e"
    end
    gsub!(/ \e/, "\e")
    gsub!(/\e+/, "\n")
    print
    @col = 2 * @tab
  end

  def add(s)
    @stac.push s
    @col += s.length
  end

  def fix
    for i in 1...(@stac.length)
       pre = @stac[i-1]
       cur = @stac[i]
       if cur == ' '
         case pre
         when '%c:', '%v:'
           @stac[i-1], @stac[i] = cur, pre
         end
       end
    end
    str = ""
    @stac.each { |s| str << s }
    @stac = []
    str
  end

  def put_flat(flat='')
    if flat.empty?
      return
    end

    flat.gsub!(/&lt;/i, '<')
    flat.gsub!(/&gt;/i, '>')
    flat.gsub!(/&amp;/i, '&')

    if @in_pre
      add flat
      line_feed
      return
    end

    if @in_tag
      flat.gsub!(/%/, '%%')
    end

    flat.split.each do |word|
      unless @stac.empty?
        add ' '
      end
      if word.length == word.split(//).length
        if @col + word.length > @maxcol
          line_feed
        end
        add word
      else
        word.scan(/(.)/) do |char,|
          if @col + char.length > @maxcol
            line_feed
          end
          add char
        end
      end
    end
  end

  def put_name(name,attr)
    line_feed if @name == name
    @name = name
    case name
    when 'pre'
      case attr['class']
      when 'rb'
        add '%ruby:'
      when 'c'
        add '%c:'
      else
        add '%r:'
      end
      @in_pre = true
    when '/pre'
      @in_pre = false
    when 'code'
      add ' ' if @stac[-1] == '%'
      add '%c:'
      @in_tag = true
    when 'var'
      add ' ' if @stac[-1] == '%'
      add '%v:'
      @in_tag = true
    when '/code','/var'
      add '%'
      @in_tag = false
    when 'title'
      add 'title:'
    when  'hr'; line_feed; @in_dd = false
    when %r!^h([1-6])!; line_feed; add '1.' * $1.to_i; @in_dd = false
    when %r!/h([1-6])!; line_feed
    when  'p' ; line_feed
    when '/p' ; line_feed
    when  'ul'; line_feed; shift
    when '/ul'; line_feed; unshift
    when  'ol'; line_feed; shift
    when '/ol'; line_feed; unshift
    when  'dl'; line_feed; shift
    when '/dl'; line_feed; @in_dd = false; unshift
    when  'dt'; line_feed; @in_dd = false
    when  'dd'; line_feed; @in_dd = true
    when  'li'; line_feed; @in_dd = false
    end
  end
end

plain = Plain.new

filename = ARGV.shift || 'foo.html'
unless test ?f, filename
  print "usage: ruby #$0 html_file\n"
  exit
end
file = open filename, 'r'
begin
  text = file.read
ensure
  file.close
end

$_ = text
while not (~/<([^>]+)>/).nil?
  flat = $`
  rest = $'
  tag = $1
  if plain.in_pre
    flat.gsub!(/\n/, "\e")
  else
    flat.strip!
    flat.gsub!(/\n/, ' ')
    flat.gsub!(/ +/, ' ')
  end
  plain.put_flat flat
  if tag[0] == ?!
    $_ = rest
    next
  end
  name, *params = tag.split
  name.downcase!
  attr = {}
  params.each { |param|
    key, val = param.split('=')
    begin
      key.downcase!
    rescue
      p $!
      p tag
      exit
    end
    val = val.sub(/^"/,'').sub(/"$/,'') unless val.nil?
    attr[key] = val
  }
  plain.put_name name,attr
  $_ = rest
end

plain.line_feed
